import os

from github import Github, GithubException
from github.GithubException import UnknownObjectException

from .linkify import linkify

DEFAULT_COLOUR = "b6d7a8"


class GithubRepo:
    def __init__(self, reponame, ui):
        token = os.environ.get("GITHUB_ACCESS_TOKEN", None)
        if token is None:
            ui.error(
                "GITHUB_ACCESS_TOKEN not set. See: <https://github.com/settings/tokens>"
            )
        try:
            g = Github(token)
            self.repo = g.get_repo(reponame)
        except GithubException as err:
            ui.error(err.data["message"], "GitHub Repo")

    def create_issue(self, title, content, labels, cc):
        _labels = []
        for label_text in labels:
            try:
                label = self.repo.get_label(label_text)
            except UnknownObjectException:
                label = self.repo.create_label(label_text, DEFAULT_COLOUR)
            _labels.append(label)
        if cc:
            content = f"_Comment by @{cc}_\n\n{content}"
        issue = self.repo.create_issue(
            title=normalise_ws(title), body=content, labels=_labels
        )
        return issue.number


def create_issues(reponame, ui, base, comments, labels, cc=None, start_num=None):
    repo = GithubRepo(reponame, ui)
    if labels is None:
        labels = []
    for num, comment in enumerate(comments, start=1):
        if start_num and num < start_num:
            continue
        title = f"{comment[0]}"
        content = linkify(comment[1], base)
        try:
            number = repo.create_issue(title, content, labels, cc)
        except GithubException as err:
            ui.error(
                f"{err.data['message']} - restart with --start={num}",
                "Github Issue Creation",
            )
        ui.status(f"* Created issue {number} in {reponame}", f"{title}")


def normalise_ws(text):
    return " ".join(text.split())



import argparse
import sys

from . import __version__
from .md_comments import parse_markdown_comments
from .xml_comments import parse_xml_comments
from .github import create_issues

from colorama import Fore, Style


class Cli:
    @classmethod
    def out(cls, content):
        sys.stdout.write(content)

    @classmethod
    def status(cls, name, value):
        sys.stderr.write(f"{Fore.GREEN}{name}:{Style.RESET_ALL} {value}\n")

    @classmethod
    def warn(cls, message, source=""):
        if source:
            source = f"{source} "
        sys.stderr.write(f"{Fore.YELLOW}{source}Warning{Style.RESET_ALL}: {message}\n")

    @classmethod
    def error(cls, message, source=""):
        if source:
            source = f"{source} "
        sys.stderr.write(f"{Fore.RED}{source}Error{Style.RESET_ALL}: {message}\n")
        sys.exit(1)

    def comment(self, comment):
        self.out(f"{Fore.BLUE}## {comment[0]}{Style.RESET_ALL}\n")
        self.out(f"{comment[1]}\n")
        self.out("\n")


def base_arg_parser(description):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-g",
        "--github_repo",
        metavar="owner/repo",
        dest="github_repo",
        help="create issues in the named repo",
    )

    parser.add_argument(
        "-l",
        "--github-label",
        dest="github_label",
        action="append",
        help="label to assign to created GitHub issues",
    )

    parser.add_argument(
        "-s",
        "--start",
        dest="start_num",
        type=int,
        metavar="NN",
        default=None,
        help="Comment number to skip to when creating issues",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{__version__}",
        help="print version and exit",
    )
    return parser


def ietf_comments_cli():
    args = parse_ietf_args()
    cli = Cli()
    comments = parse_markdown_comments(args.comment_file, cli)
    base = f"https://www.ietf.org/archive/id/{comments.doc}-{comments.revision}.html"
    cli.status(f"Document", comments.doc)
    cli.status(f"Revision", comments.revision)
    if comments.cc:
        cli.status(f"CC", f"@{comments.cc}")
    for issue_type in comments.sections:
        these_comments = comments.issues[issue_type]
        if args.github_repo:
            if args.auto_label:
                labels = ["review", issue_type]
            else:
                labels = args.github_label or []
            create_issues(
                args.github_repo,
                cli,
                base,
                these_comments,
                labels,
                comments.cc,
                args.start_num,
            )
        else:
            cli.out(f"\n{Fore.GREEN}# {issue_type}{Style.RESET_ALL}\n")
            for comment in these_comments:
                cli.comment(comment)


def parse_ietf_args():
    parser = base_arg_parser("Process a markdown file containing IETF comments")
    parser.add_argument(
        "-a",
        "--auto-label",
        dest="auto_label",
        action="store_true",
        help="Add a 'review' label, and 'discuss', 'comment', or 'nit' as appropriate",
    )
    parser.add_argument(
        "comment_file",
        metavar="filename",
        type=argparse.FileType("r"),
        help="comment file to process",
    )
    return parser.parse_args()


def rfced_comments_cli():
    args = parse_rfced_args()
    cli = Cli()
    comments = parse_xml_comments(args.rfc, cli)
    if args.rfc.isnumeric():
        rfcnum = args.rfc
    else:
        rfcnum = "NNNN"
    if args.github_repo:
        labels = args.github_label or []
        base = f"https://www.rfc-editor.org/authors/rfc{rfcnum}.html"
        create_issues(
            args.github_repo, cli, base, comments, labels, start_num=args.start_num
        )
    else:
        for comment in comments:
            cli.comment(comment)


def parse_rfced_args():
    parser = base_arg_parser("Process an AUTH48 draft for RFC Editor comments")
    parser.add_argument(
        "rfc",
        metavar="NNNN",
        help="RFC-to-be number or filename to process",
    )
    return parser.parse_args()



import re
from io import StringIO
import os
import textwrap
import xml.sax

import requests


def parse_xml_comments(rfc, ui):
    if not os.path.isfile(rfc):
        rfcxml = fetch_rfcxml(rfc, ui)
        rfc = StringIO(rfcxml)
    parser = xml.sax.make_parser()
    handler = XmlCommentHandler()
    parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler)
    parser.parse(rfc)
    return handler.comments


class XmlCommentHandler(xml.sax.handler.LexicalHandler):
    source_regex = re.compile(
        r"^\s*\[\s*([a-zA-z0-9\-]+)\s*\]\s*(.+)", flags=re.MULTILINE | re.DOTALL
    )
    colon_title = re.compile(r"^([^:]{,70}):")
    question_title = re.compile(r"^([^?]{,70}\?)", re.MULTILINE)
    sentence_title = re.compile(r"^(.{,70})(?=\.\s+)", re.MULTILINE)
    quoted_start = re.compile(
        r"(Currently|Original|Suggested|Possibly):\s*\n(\s+.*?)(\n\n|$)", re.DOTALL
    )

    def __init__(self):
        self.comments = []

    def comment(self, content):
        result = self.source_regex.match(content)
        if result:
            source = result.group(1)
            if source == "rfced":
                raw_comment = result.group(2)
                title = self.extract_title(raw_comment)
                comment = self.process_comment(raw_comment)
                self.comments.append((title, comment))

    def extract_title(self, text):
        colon_result = self.colon_title.match(text)
        comment_num = len(self.comments) + 1
        comment_counter = f" (comment {comment_num})"
        if colon_result:
            return colon_result.group(1).strip() + comment_counter
        question_result = self.question_title.match(text)
        if question_result:
            return question_result.group(1).strip() + comment_counter
        sentence_result = self.sentence_title.match(text)
        if sentence_result:
            return sentence_result.group(1).strip() + comment_counter
        return f"RFC Editor comment {comment_num}"

    def process_comment(self, text):
        adjusted = self.quoted_start.sub(self.blockquote_indent, text)
        return adjusted

    @classmethod
    def blockquote_indent(cls, match):
        indented = textwrap.indent(match.group(2), "    ", lambda line: True)
        return f"{match.group(1)}:\n\n{indented}{match.group(3)}"


def fetch_rfcxml(rfcnum, ui):
    if not rfcnum.isnumeric():
        ui.error(f"'{rfcnum}' is not a file and not numeric.")
    url = f"https://www.rfc-editor.org/authors/rfc{rfcnum}.xml"
    res = requests.get(url)
    if res.status_code != 200:
        ui.error(f"RFC{rfcnum}-to-be not found on RFC Editor server.")
    return res.text



import re


flags = re.IGNORECASE | re.MULTILINE | re.VERBOSE
ref = r"[a-zA-Z0-9\-_]+"
section_finder = re.compile(
    rf"""
(?:
    \[(?P<comma_ref>{ref})\],\s*
)?
(?P<section>(?:section|sections|appendix|§)\s+)
(?P<num>[\d\.]*\d)\.?
(?:
    \s+of\s+\[(?P<of_ref>{ref})\]
)?
""",
    flags,
)


def linkify(text, spec_uri):
    def linker(matchobj):
        section = matchobj.group("section")
        if section.lower().strip() == "appendix":
            sref = "appendix"
        else:
            sref = "section"
        num = matchobj.group("num")
        if matchobj.group("comma_ref"):
            ref = matchobj.group("comma_ref")
            link = find_ref_uri(ref)
            if link:
                return f"[{ref}]({link}), [{section}{num}]({link}#{sref}-{num})"
            else:
                return f"[{ref}], {section}{num}"
        elif matchobj.group("of_ref"):
            ref = matchobj.group("of_ref")
            link = find_ref_uri(ref)
            if link:
                return f"[{section}{num}]({link}#{sref}-{num}) of [{ref}]({link})"
            else:
                return f"{section}{num} of [{ref}]"
        else:
            return f"[{section}{num}]({spec_uri}#{sref}-{num})"

    return section_finder.sub(linker, text)


def find_ref_uri(ref):
    if ref[:3].lower() == "rfc" and ref[3:].isnumeric():
        return f"https://rfc-editor.org/rfc/rfc{ref[3:]}.html"


if __name__ == "__main__":
    print(linkify("in section 4.3 of that", "https://www.example.com/"))
    print(linkify("in section 4.3. of that", "https://www.example.com/"))
    print(linkify("in [RFC3221], section 4.3 of that", "https://www.example.com/"))
    print(linkify("in Section  4.3 of [RFC2119]", "https://www.example.com/"))
    print(linkify("in section 4.3 of [OTHER_THING]", "https://www.example.com/"))
    print(linkify("in appendix 4.3 of [RFC1234]", "https://www.example.com/"))
    print(linkify("in section 4.3a of [OTHER_THING]", "https://www.example.com/"))
    print(linkify("in section 4a.3a of [OTHER_THING]", "https://www.example.com/"))
    print(linkify("in section a.3a of [OTHER_THING]", "https://www.example.com/"))



import commonmark
from commonmark.render.renderer import Renderer

from .changes import DocChanges


class CommentRenderer(Renderer):
    sections = ["discuss", "comment", "nit"]
    section_map = {"discusses": "discuss", "comments": "comment", "nits": "nit"}
    BLOCK = "block"

    def __init__(self, ui, options={}):
        self.ui = ui
        self.options = options
        self._buffer = []
        self._section = None
        self._current_issue = None
        self._context = None
        self._context_buffer = []
        self.title = None
        self.doc = None
        self.revision = None
        self.cc = None
        self.changes = None
        self.issues = {"discuss": [], "comment": [], "nit": []}

    def text(self, node, entering=None):
        self._buffer.append(node.literal)
        if self._context is not None:
            self._context_buffer.append(node.literal)

    def softbreak(self, node, entering=None):
        self._buffer.append(" ")
        if self._context_buffer:
            self._context_buffer.append(" ")

    def get_buffer(self):
        content = "".join(self._buffer)
        self._buffer = []
        return content

    def heading(self, node, entering):
        if entering:
            self.cleanup()
        else:
            content = self.get_buffer().lower()
            getattr(self, f"handle_h{node.level}")(content)

    def handle_h1(self, content):
        if self.doc is not None:
            self.ui.error("More than one h1 header.")
        self.title = content
        docname = None
        words = content.split()
        for word in words:
            if word.startswith("draft-"):
                docname = word
                break
        if docname is None:
            self.ui.error(
                "h1 header doesn't contain draft name (starting with 'draft-')."
            )
        segments = docname.split("-")
        self.doc = "-".join(segments[:-1])
        revision = segments[-1]
        if revision.isnumeric():
            self.revision = revision
        else:
            self.ui.error("h1 header draft name doesn't include revision")
        self.changes = DocChanges(self.doc, self.revision, self.ui)

    def handle_h2(self, content):
        content = content.lower()
        if content in self.section_map:
            content = self.section_map[content]
        if content in self.sections:
            self._section = content
        else:
            self.ui.warn(f"Unrecognised h2 section {content}.")
            self._section = None

    def handle_h3(self, content):
        self._current_issue = content

    def item(self, node, entering):
        if entering:
            self._buffer.append("* ")

    def paragraph(self, node, entering):
        if not entering:
            self._buffer.append("\n\n")

    def block_quote(self, node, entering):
        if entering:
            self._buffer.append("> ")
            self._context = self.BLOCK
        else:
            self._context = None
            content = "".join(self._context_buffer)
            self._context_buffer = []
            change_location = self.changes.find_change_line(content)
            if change_location is None:
                self.ui.warn(f"Can't find quoted text in document: {content}")

    def emph(self, node, entering):
        self._buffer.append("_")

    def strong(self, node, entering):
        self._buffer.append("**")

    def link(self, node, entering):
        if entering:
            self._buffer.append("[")
        else:
            self._buffer.append(f"]({node.destination})")

    def document(self, node, entering):
        if not entering:
            self.cleanup()
            if sum([len(l) for l in self.issues.values()]) == 0:
                self.ui.warn("Did not find any issues.")

    def cleanup(self):
        if self._section:
            if self._current_issue is None:
                self._current_issue = self.title
            text = self.get_buffer()
            if text:
                self.issues[self._section].append((self._current_issue, text))
        else:
            self.process_preface()
        self._current_issue = None
        self._buffer = []

    def process_preface(self):
        text = self.get_buffer()
        if text:
            for line in text.split("\n"):
                if line.lstrip().lower().startswith("cc @"):
                    self.cc = line.strip().split("@", 1)[1]
                    break


def parse_markdown_comments(fd, ui):
    parser = commonmark.Parser()
    doc = parser.parse(fd.read())
    renderer = CommentRenderer(ui)
    renderer.render(doc)
    return renderer



import requests


class DocChanges:
    URL_BASE = "https://www.ietf.org/archive/id/"

    def __init__(self, docname, revision, ui):
        self.docname = docname
        self.revision = revision
        self.ui = ui
        self.text_doc = self.get_doc("txt")

    def get_doc(self, doctype):
        res = requests.get(f"{self.URL_BASE}/{self.docname}-{self.revision}.{doctype}")
        if res.status_code != 200:
            self.ui.warn(
                f"Can't find {self.docname}-{self.revision} text on IETF servers."
            )
        doc = res.text
        return doc

    def find_change_line(self, old):
        old_words = old.split()
        old_word_count = len(old_words)
        line_no = 0
        lines_consumed = 0
        words_consumed = 0
        for line in self.text_doc.split("\n"):
            line_no += 1
            line_words = line.split()
            line_word_count = len(line_words)
            line_words_consumed = 0
            if old_words[0] in line_words or words_consumed > 0:
                if words_consumed > 0:
                    start_word = 0
                else:
                    try:
                        start_word = line_words.index(old_words[0])
                    except ValueError:
                        continue
                for line_word in line_words[start_word:]:
                    if line_word != old_words[words_consumed]:
                        words_consumed = lines_consumed = 0
                        break
                    line_words_consumed += 1
                    words_consumed += 1
                    if words_consumed == old_word_count:
                        return line_no - lines_consumed, lines_consumed + 1
                if line_words_consumed + start_word == line_word_count:
                    lines_consumed += 1
